/* 
 * 
 *     FiPaBoL for linux-7110
 *     (C)   R G Gammans    
 * 
 *     19 October 1998
 * 
 *     v0.1 (no RCS yet)
 *
 *     Badly hacked by Werner Almesberger
 *
 *     This code is licensed under the GPL
 *      - Hey. It's linux what else did you expect.
 */


/* 
 *   This code copies and concatenates a list of pages passed to it
 *   in a page pointed to by R0.
 *   The table is terminated by a 0.
 * 
 *   The destination address is first word in the second page.
 * 
 *   The first page is ignored as assumed to point to us.
 *   The second page of the table is assumed to be ws.
 *   Eg we prepend the kernel bootimage with this page 
 *      then an extra empty page.
 * 
 *   The code never exits just jumps to destination address.
 * 
 *   It fetches the dest address of the page before copy, then scans
 *   the list of sources (And its own location). It relocates the
 *   destination page if it would be overwritten to a pre-determined
 *   free page updating the table list as appropriate.
 * 
 *   After a page is copied it src location become the current free
 *   page. The empty extra page in the boot file is used as the
 *   first 'free' page.
 */


/*
 * Register usage:
 *
 * R0	pointer to current page table entry
 * R1	destination address
 * R2	source address
 * R3	pointer to beginning of page table
 * R4	entry we compare with in table scans
 * R5	pointer to code page
 * R6	temporary local storage (during calls to copypage)
 * R7	clobbered by copypage & debug
 * R8	pointer into the free memory list (points to next fragment)
 * R9	amount of free memory remaining in current block
 * R10	clobbered by copypage & debug
 * R11	ditto
 * R12  pointer to page table entry when searching for conflicts
 * R13	scrap space table
 *
 * Input registers: R0, R5, R8, LR
 */
	.align
	.text

pagesize	=	4096

entry:	@  Startup
#ifndef	R5_IS_CODE_PAGE
	mov	r5,pc,lsr #12		@ Store page address of
	mov	r5,r5,lsl #12		@ code page...
#endif


	@ The next few instructions allow us to recover
	@ control flow back to our test harness to check
	@ the correct operation of the bootloader.
	@ (And hide any cpsr flags which have found their
	@  way into R5 if we test in 26 bit mode.).
	@ 
	@ This trick requires return code though as part
	@ the defragged image.

#ifdef Debug
#ifdef ARM26
	bics	r5,r5,#0xfc000000
#endif
	add	r6,r5,#4000		@ Some stack space...
	stmia	r6!,{r9-r14}
#endif

	@ Now the first address in the page table should point to us
	@ the second will point to our 'scrap' space...

	mov     r3,r0			@ Store page address of pagetable
	ldr     r13,[r0,#12]!		@ Get pre-inc word from table 
					@  eg. our scrap space..
	
	@ These two instructions move our destination address
	@  so I can keep it safe during defrag..
	@  Or you could just hard code it in at destjump
	@  and remove these two.	

#ifndef STANDALONE
	ldmia   r8!,{r1,r9}		@ load start and length of fragment
	add	r1,r1,#KERNEL_OFFSET	@ adjust ...
	sub	r9,r9,#KERNEL_OFFSET
#else
	ldr	r1,[r13]		@ Get our destination address.
here:
	str 	r1,[pc,#(destjump-here-8)]
					@ Ensure PC relative
					@ even though most asms gen this way
#endif

loop1:					@ entry with pre-loaded regs
	dbgbeg	r11
	dbgchr	#'.',r11,r7
	dbgend	r11,r7

	ldr	r2,[r0,#4]!		@ Fetch next item

	dbgbeg	r11
	dbgreg	#'C',r5,r11,r7,r10	@ code page
	dbgreg	#'P',r3,r11,r7,r10	@ page table
	@dbgreg	#'p',r0,r11,r7,r10	@ page pointer
	dbgreg	#'S',r2,r11,r7,r10	@ source
	dbgreg	#'D',r1,r11,r7,r10	@ destination
	@dbgreg	#'T',r13,r11,r7,r10	@ tmp reg
	dbgend	r11,r7

	cmp     r2,#0
	beq	exit

	@  Scan pages for clash..

	mov	r12,r0			@ Duplicate current table pos.

scanloop:
	ldr     r4,[r12,#4]!		@ Get page addr from table
	cmp     r4,#0			@ EOF ?
	ldreq	r4,[r12,#4]!		@ get next page (if any)
	cmpeq	r4,#0			@ EOT (end of table) ?
	beq	endscan			@ jump out of loop
	cmp     r4,r1			@ Is this our current dest page
	beq     relocpage		@ if p<>destpage then
	b	scanloop		@   goto loop

endscan:
	@  copy to relocate page or self if necessary

	@ check codepage ok
	cmp	r5,r1			@   if dest=codepage then
	bleq	relocself		@      relocself

	@ check table page ok
	cmp     r3,r1			@   if dest=table
	bne	donereloc		@     then

	mov     r12,r1			@     store destination
	mov     r1,r13			@     setup regs to copy..
	mov	r2,r3			@     table to temp page
	sub	r0,r0,r3		@     Make R0 offset into table
	mov     r3,r1			@     Keep track of page table
	add     r0,r0,r3		@     .. and our pointer into it
	bl      copypage		@     do it

	mov     r1,r12			@     restore image destination
					@  (end) else relocpage
donereloc:
	ldr     r2,[r0]			@ get table entry again

	cmp     r1,r2			@ Is there anything to do ?
	addeq	r1,r1,#pagesize		@   no -> handle next page
	beq	page_done

	bl	copypage		@ if so do it and
	sub	r13,r2,#pagesize	@   Get temp page address

page_done:
	subs	r9,r9,#pagesize		@ need new fragment ?
	ldmeqia	r8!,{r1,r9}		@ yes, get start and length
					@ @@@ reboot on error
	b       loop1			@ if not goto loop1

	@   FIXME:
	@   If running on an SA / Harvard cached CPU
	@   need to flush I-cache here.
	@ 

#ifdef STANDALONE

exit:
	ldr     pc,[pc,#-4]		@ Jump to copied image
destjump:
	.word	0

#endif

relocpage:
	@ calls copy page, then updates the table of source pages
	@ appropriately
	@ must preserve r0,r1,r3
	@  R4  contains pointer to page to relocate
	@  R12 contains pointer to page table which will need updating


	mov	r6,r1
	mov     r1,r13
	mov     r2,r4
	bl      copypage
	str     r13,[r12]
	mov     r1,r6
	@ return job done
	b       donereloc

relocself:
	@ Calls copy page then modifies R14 so returns to relocated
	@ copy of self. Cannot be called from a subroutine! As then
	@  that subroutines return adress will become invalid
	@ must preserve r0,r1,r3
	@ r5 contains pointer to base of code page


	mov	r12,r1			@ save r1
	mov	r6,lr			@ save 'return address' (ish)
	mov	r2,r5			@ copy us
	mov	r1,r13			@ copy code to temp..
	sub	r8,r8,r5		@ make R8 offset into code page
	mov	r5,r1			@ Set r5 to point to new codepage
	add	r8,r8,r5		@ ... and update pointer into it
	bl	copypage
	sub	r2,r1,r2		@ Compute difference between pages
	mov	r1,r12			@ Restore R1..

	@   FIXME:
	@   If running on an SA / Harvard cached CPU
	@   need to flush I-cache here.
	@ 
	add	pc,r6,r2		@ Compute and goto return code
					@ in new page!
copypage:
	@ on entry:-
	@   R1- Destination address
	@   R2- source addres
	@ on exit
	@   R1 - Dest address+pagesize
	@   R2 - source address+pagesize
	@   R10-11 corrupted.

	dbgbeg	r11
	dbgreg	#'>',r2,r11,r7,r10
	dbgreg	#'-',r1,r11,r7,r10
	dbgend	r11,r7

	add     r7,r2,#pagesize		@ Compute last copy addr+1
copyloop:
	ldmia   r2!,{r10-r11}
	cmp     r2,r7			@ Obscure order for speedup
					@ on some ARM family members
	stmia	r1!,{r10-r11}
	blt	copyloop
	dbgbeg	r11
	dbgreg	#'=',lr,r11,r7,r10
	dbgend	r11,r7
	mov	pc,lr

#ifdef STANDALONE
	.end
#endif
