/*
 * boot.S - Boot header for the Linux kernel Image
 *
 * Written 1998,1999 by Werner Almesberger
 */

/*
 * We first move the parameters (2kB at 0x800-0xfff relative to our load
 * address) to 0xc0023000, relocate the kernel to 0xc0038000, and jump to that
 * address, where we should rendez-vous with head-armv.S
 *
 * All this is glued together by a script, so the assembler code doesn't tell
 * half of the truth. Also, we don't probe hardware but rely exclusively on
 * the wisdom of the script configurator.
 */

/*
 * Most of these parameters already exist in asm-arm/arch-7110/, but we copy
 * them anyway (for now) in order to be self-contained. @@@
 */

#define RAM_START		0xc0000000
#define	PARAM_SRC_OFFSET	0x00000800
#define FREELIST_OFFSET		0x00000e00
#define PARAM_DST		0x00023000
#define	KERNEL_OFFSET		0x00038000
#define PAGE_SIZE		4096
#define FREELIST_SIZE		512

#define INITRD_START_OFFSET	64
#define	BOOT_OPTIONS_OFFSET	1280

#define	IO_START		0x80000000
#define SYSCON			0x0100
#define SYSFLG			0x0140
#define UARTDR			0x0480
#define UBRLCR			0x04c0
#define UBUSY			0x00000800
#define URXFE			0x00400000
#define UTXFF			0x00800000
#define	UARTEN			0x00000100
#define	SIREN			0x00008000
#define BR_9600			23
#define WRDLEN_SHIFT		17
#define WL_8			0x3
#define PCDR			0x0002
#define PSION_SER_ENABLE	0x09 /* activate drivers, assert RTS, no DTR */

#ifdef DEBUG

	.macro	dbgbeg,iobreg
	mov	\iobreg,#IO_START
	.endm

	.macro	dbgreg,id,reg,iobreg,sc1,sc2
1:	ldr	\sc1,[\iobreg,#SYSFLG]
	tst	\sc1,#UTXFF
	bne	1b
	mov	\sc1,\id
	strb	\sc1,[\iobreg,#UARTDR]
	mov	\sc2,\reg
2:	ldr	\sc1,[\iobreg,#SYSFLG]
	tst	\sc1,#UTXFF
	bne	2b
	and	\sc1,\sc2,#15
	add	\sc1,\sc1,#'0'
	cmp	\sc1,#('0'+10)
	addge	\sc1,\sc1,#(32+7)
	strb	\sc1,[\iobreg,#UARTDR]
	movs	\sc2,\sc2,lsr #4
	bne	2b
	.endm

	.macro	dbgchr,ch,iobreg,sc
1:	 ldr     \sc,[\iobreg,#SYSFLG]
        tst     \sc,#UTXFF
        bne     1b
        mov     \sc,\ch
        strb    \sc,[\iobreg,#UARTDR]
	.endm

	.macro	dbgend,iobreg,sc
1:	ldr	\sc,[\iobreg,#SYSFLG]
	tst	\sc,#UBUSY
	bne	1b
	.endm

#else

	.macro	dbgbeg,iobreg
	.endm

	.macro	dbgreg,id,reg,iobreg,sc1,sc2
	.endm

	.macro	dbgchr,ch,iobreg,ch
	.endm

	.macro	dbgend,iobreg,sc
	.endm

#endif

/*
 * Preserve R0 until calling "entry"
 */

	.align
	.text

boot:
#if defined(DEBUG) || defined(INIT_SERIAL)
	mov	r1,#IO_START		/* initialize the UART */
	ldr	r2,[r1,#SYSCON]
	bic	r2,r2,#SIREN		/* no SIR */
	orr	r2,r2,#UARTEN		/* enable UART */
	str	r2,[r1,#SYSCON]
#ifdef PSION
	mov	r2,#PSION_SER_ENABLE	/* activate drivers, assert DTR, RTS */
	strb	r2,[r1,#PCDR]
#endif
	mov	r2,#(WL_8 << WRDLEN_SHIFT) /* 8 bits */
	orr	r2,r2,#BR_9600		/* and 9600 bps */
	str	r2,[r1,#UBRLCR]
#ifdef DEBUG
	mov	r2,#'('			/* ( */
	strb	r2,[r1,#UARTDR]
#endif
#endif

	dbgbeg	r11
	mov	r5,pc,lsr #12		/* pointer to the free memory list */
	mov	r5,r5,lsl #12
	orr	r8,r5,#FREELIST_OFFSET
	ldr	r10,[r8]		/* do we have a free memory list ? */
	cmp	r10,#0
	beq	findmem			/* no -> find free memory first */
bffm:	mov	r5,pc,lsr #12		/* restore R5 and R8 (clobbered by */
	mov	r5,r5,lsl #12		/*   findmem) */
	orr	r8,r5,#FREELIST_OFFSET
	orr	r7,r5,#PARAM_SRC_OFFSET	/* pointer to boot options */
	orr	r7,r7,#BOOT_OPTIONS_OFFSET
	ldr	r2,[r0]			/* options passed by user */
	ldrb	r1,[r2]			/* anything there ? */
	tst	r1,#0xff
	beq	entry			/* no -> skip */
optelp:	ldrb	r1,[r7],#1		/* skip to end of pre-set options */
	tst	r1,#0xff
	bne	optelp
	mov	r1,#32			/* append a space */
	strb	r1,[r7,#-1]
optcplp:ldrb	r1,[r2],#1		/* copy user-provided string */
	strb	r1,[r7],#1
	tst	r1,#0xff
	bne	optcplp
	b	entry			/* start the shuffle */

	/* plenty of black magic from FiPaBol.S ... */

exit:
	dbgbeg	r1
	dbgchr	#')',r1,r2

	ldr	r2,[r0,#4]		/* more files ? */
	cmp	r2,#0
	beq	dncpf			/* no - done copying files *

/*
 * Register usage:
 *
 * R1,R8,R9	walk the free list backwards (see FiPaBol.S)
 * R6		available space in current fragment
 * R10		pointer into file (to find EOF)
 * R11		file size
 */

	orr	r8,r5,#FREELIST_OFFSET	/* find end of free list ... */
fefllp:	ldr	r1,[r8,#8]!
	cmp	r1,#0			/* at end ? */
	bne	fefllp			/* no -> continue */

	mov	r10,r0			/* temporary file walk pointer */
	mov	r11,#0			/* file size */
	mov	r6,#0			/* need to get next descriptor */

fsbwlp:	ldr	r2,[r10,#4]!		/* get next page */
	cmp	r2,#0			/* at end ? */
	beq	fsbwdn			/* yes -> copy that file */
	add	r11,r11,#PAGE_SIZE	/* no -> increase size */
	cmp	r6,#0			/* space left in the current frag ? */
	ldmeqdb	r8!,{r1,r6}		/* no: get next descriptor */
	addeq	r1,r1,r6		/* no: pointer to the end */
	moveq	r9,#0			/* no: no bytes used yet */
	sub	r1,r1,#PAGE_SIZE	/* advance pointer (removing all those*/
	sub	r6,r6,#PAGE_SIZE	/*   counters from the loop would     */
	add	r9,r9,#PAGE_SIZE	/*   save only milliseconds ...)      */
	b	fsbwlp			/* next round */

fsbwdn:	add	r2,r5,#PARAM_SRC_OFFSET	/* register file start and size as */
	add	r2,r2,#INITRD_START_OFFSET /* initrd parameters */
	str	r1,[r2],#4
	str	r11,[r2]
	add	r8,r8,#8		/* move r8 from current to next frag */
	b	loop1			/* copy file */

dncpf:	mov	r12,#RAM_START		/* copy the parameter block */
	mov	r2,pc,lsr #12
	mov	r2,r2,lsl #12
	orr	r2,r2,#PARAM_SRC_OFFSET
	orr	r3,r12,#PARAM_DST
	mov	r4,#512			/* 512*4 bytes */
pclp:	ldr	r5,[r2],#4		/* copy one word */
	str	r5,[r3],#4
	subs	r4,r4,#1		/* next iteration */
	bne	pclp

	dbgchr	#13,r1,r2
	dbgchr	#10,r1,r2

	mov	r0,#0			/* head-armv convention */
	mov	r1,#9			/* we're a CL-PS7110 */
	mov	r12,#RAM_START		/* go ! */
	orr	pc,r12,#KERNEL_OFFSET

#undef	STANDALONE
#define	R5_IS_CODE_PAGE
#include "FiPaBol.S"

#ifdef AUTODETECT
#include "findmem.S"
#else
findmem:
	ldr	r2,[r0,#4]		/* pointer to EPOC-provided table */
	add	r2,r2,#8		/* skip first entry */
	ldmia	r2!,{r6,r7}		/* second entry is special ... */
	sub	r6,r6,#RAM_START	/* we know where the RAM starts, so */
	add	r7,r7,r6		/*   we include all memory reserved */
	mov	r6,#RAM_START		/*   by EPOC */
	stmia	r8!,{r6,r7}
emtclp:	dbgreg	#'@',r2,r11,r12,r13
	dbgreg	#'>',r8,r11,r12,r13
	ldmia	r2!,{r6,r7}		/* copy one entry */
	stmia	r8!,{r6,r7}
	dbgreg	#'=',r6,r11,r12,r13
	dbgreg	#'+',r7,r11,r12,r13
	cmp	r6,#0			/* at end ? */
	bne	emtclp			/* no -> next entry */
	b	bffm			/* done */
#endif

	.end
