#!/usr/bin/perl
#
# glue.pl - glue together a load image
#
# The ingredients:
#   boot.bin	boot setup code (will include FiPaBoL)
#   params	data for the parameter block
#   freelist	list of available memory regions
#   Image	guess what :)
#
# The result:
#   stdout	run and enjoy it
#
$BOOT = "boot.bin";
$PARAMS = "params";
#$FREELIST = "freelist";
$IMAGE = "Image";

#
# Currently these three *must*
# sum to 4096 - the size of one page.
#

$BOOTSIZE =  2048;
$PARAMSIZE = 1536;
$FREESIZE  = 512;

die "internal error" unless $BOOTSIZE+$PARAMSIZE+$FREESIZE == 4096;

sub usage
{
    print STDERR "usage: $0 [ -b boot ] [ -p params ] [ -f freelist ] ".
      " [ -o outfile ] [ image ]\n";
    exit 1;
}

while ($_ = shift @ARGV) {
    if (/^-/) {
	$next = shift @ARGV;
	&usage unless defined $next;
	if ($_ eq "-b") { $BOOT = $next; }
	elsif ($_ eq "-p") { $PARAMS = $next; }
	elsif ($_ eq "-f") { $FREELIST = $next; }
	elsif ($_ eq "-o") { $OUTFILE = $next; }
	else { &usage; }
	next;
    }
    $arg++;
    if ($arg == 1) {
	$IMAGE = $_;
    }
    elsif ($arg == 2) {
	print STDERR "initrd is no longer merged with the kernel image\n";
	exit 1;
    }
    else {
	&usage;
    }
}

print STDERR "Creating boot image".(defined $OUTFILE ? " $OUTFILE" : "").":\n";
open(F,$BOOT) || die "$BOOT: $!";
$size = (stat F)[7];
sysread(F,$image,$size) == $size || die "$BOOT: $!";
close F || die "$BOOT: $!";
die "$BOOT is too large ($size > $BOOTSIZE)" if $size > $BOOTSIZE;
print STDERR "  Relocator  $BOOT($size)\n";

$image .= "\0" x ($BOOTSIZE-$size);

open(F,$PARAMS) || die "$PARAMS: $!";
while (<F>) {
    chop;
    next if /^\s*(|#.*)$/;
    if (/^(\.string)\s+"([^"]*)"\s*(|#.*)$/) {
	$param .= $2;
	next;
    }
    elsif (/^(\.skip)()\s+(\d+|0x[0-9a-fA-Z]+)\s*(|#.*)$/) { }
    elsif (/^(\.8)()\s+\S+\s+(\d+|0x[0-9a-fA-Z]+)\s*(|#.*)$/) { }
    elsif (/^()(\S+)\s+(\d+|0x[0-9a-fA-Z]+)\s*(|#.*)$/) { }
    else {
	die "unrecognized: $_";
    }
    $value = substr($3,0,2) eq "0x" ? hex($3) : $3;
    if ($1 eq ".skip") {
	$param .= "\0" x $value;
	next;
    }
    if ($1 eq ".8") {
	$param .= pack("C",$value);
	next;
    }
    $param .= pack("i",$value);
}
die "Params too large ".length($param)." > ".$PARAMSIZE if
  length($param) > $PARAMSIZE;
print STDERR "  Parameters $PARAMS(".length($param).")\n";
$image .= $param.("\0" x ($PARAMSIZE-length($param)));

if (defined $FREELIST) {
    open(F,$FREELIST) || die "$FREELIST: $!";
    while (<F>) {
	chop;
	next if /^\s*(|#.*)$/;
	die "unrecognized: $_" unless /^\s*([0-9a-fA-F]+)\s+([0-9a-fA-F]+)\s*$/;
	$list .= pack("II",hex($1),hex($2));
    }
    close F;
    die "Freelist too large ".length($list)." > ".$FREESIZE if
      length($list) > $FREESIZE;
    print STDERR "  Freelist   $FREELIST(".length($list).")\n";
}
$image .= $list.("\0" x ($FREESIZE-length($list)));

# add scratch page
$image .= "\0" x 4096;

open(F,$IMAGE) || die "$IMAGE: $!";
$size = (stat F)[7];
sysread(F,$tmp,$size) == $size || die "$IMAGE: $!";
close F || die "$IMAGE: $!";
print STDERR "  Kernel     $IMAGE($size)\n";
$image .= $tmp;

$image .= ("\0" x (4096-(length($image) & 4095)));
print STDERR "Boot image: ".length($image)." bytes\n";

if (!defined $OUTFILE) {
    print $image || die "(stdout): $!";
}
else {
    open(FILE,">$OUTFILE") || die "create $OUTFILE: $!";
    print FILE $image || die "write $OUTFILE: $!";
    close FILE || die "close $OUTFILE: $!";
}
