#!/usr/bin/perl
#
# mergemem.pl - take lists of memory areas, merge them, and compute the
#		smallest step size to cover all areas
#
for $file (@ARGV) {
    open(FILE,$file) || die "$file: $!";
    while (<FILE>) {
	chop;
	s/\s*#.*//;
	s/^\s*//;
	s/\s+/ /g;
	s/\s*$//;
	next if /^$/;
	die "unrecognized: $_" unless /^([0-9A-Fa-f]+) ([0-9A-Fa-f]+)$/;
	push(@in_start,hex($1));
	push(@in_size,hex($2));
    }
}
for $i (0..$#in_start) {
    next unless $in_start[$i];
    local ($start,$end);

    $start = $in_start[$i];
    $end = $start+$in_size[$i];
    $in_size[$i] = 0;
    &constraint($in_start[$i]);
    for ($j = 0; $j <= $#in_start; $j++) {
	next unless $in_start[$j];
	next if $end < $in_start[$j];
	next if $in_start[$j]+$in_size[$j] < $start;
	$tmp = $start < $in_start[$j] ? $start : $in_start[$j];
	&constraint($in_size[$j]);
	&constraint($start-$tmp);
	&constraint($in_start[$j]-$tmp);
	$end = ($end > $in_start[$j]+$in_size[$j] ? $end :
	  $in_start[$j]+$in_size[$j]);
	$start = $tmp;
	$in_start[$j] = 0;
	$j = 0;
    }
    push(@out_start,$start);
    push(@out_size,$end-$start);
}
print "/*\n * Machine-generated file. Do not edit.\n */\n".
  "#define\tSTEP_SIZE\t$constraint\n" || die "print: $!";
for $i (sort { $out_start[$a] <=> $out_start[$b] } (0..$#out_start)) {
    printf("\t.long\t0x%08x\t\t/* %d step%s */\n\t.long\t0x%08x\n",
      $out_start[$i],$out_size[$i]/$constraint,
      $out_size[$i] == $constraint ? "" : "s",$out_size[$i])
      || die "print: $!";
    $steps += $out_size[$i]/$constraint;
}
print "#define\tSTEPS\t$steps\n" || die "print: $!";


sub constraint
{
    local ($new) = @_;

    return unless $new;
    $constraint = defined($constraint) ? &gcd($new,$constraint) : $new;
}


sub gcd
{
    local ($a,$b) = @_;

    while ($a != $b) {
	if ($a > $b) {
	    $a = $a-$b;
	}
	else {
	    $b = $b-$a;
	}
    }
    return $a;
}
